//----------------------------------------------------------------------------
//
// File:        main.cpp
// Date:        02-Apr-2000
// Programmer:  Marc Rousseau
//
// Description: This file contains starup code for Linux/SDL
//
// Copyright (c) 2000-2001 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.hpp"
#include "logger.hpp"
#include "cartridge.hpp"
#include "SDL/SDL.h"
#include "tms9900.hpp"
#include "ti994a.hpp"
#include "ti994a-sdl.hpp"
#include "tms9918a.hpp"
#include "tms9918a-sdl.hpp"
#include "tms9919.hpp"
#include "tms9919-sdl.hpp"
#include "tms5220.hpp"
#include "device.hpp"
#include "diskio.hpp"
#include "ti-disk.hpp"
#include "fileio.hpp"
#include "option.hpp"

DBG_REGISTER ( __FILE__ );

static sRGBQUAD ColorTable [2][17] = {
    {
        { 0x00, 0x00, 0x00, 0x00 },		// 0x00 TI_TRANSPARENT - Background
        { 0x00, 0x00, 0x00, 0x00 },		// 0x01 TI_BLACK
        { 0x00, 0xCC, 0x00, 0x00 },		// 0x02 TI_MEDIUM_GREEN
        { 0x00, 0xFF, 0x00, 0x00 },		// 0x03 TI_LIGHT_GREEN
        { 0x00, 0x00, 0x80, 0x00 },		// 0x04 TI_DARK_BLUE
        { 0x00, 0x00, 0xFF, 0x00 },		// 0x05 TI_LIGHT_BLUE
        { 0x80, 0x00, 0x00, 0x00 },		// 0x06 TI_DARK_RED
        { 0x00, 0xFF, 0xFF, 0x00 },		// 0x07 TI_CYAN
        { 0xCC, 0x00, 0x00, 0x00 },		// 0x08 TI_MEDIUM_RED
        { 0xFF, 0x00, 0x00, 0x00 },		// 0x09 TI_LIGHT_RED
        { 0xB0, 0xB0, 0x00, 0x00 },		// 0x0A TI_DARK_YELLOW
        { 0xFF, 0xFF, 0x00, 0x00 },		// 0x0B TI_LIGHT_YELLOW
        { 0x00, 0x80, 0x00, 0x00 },		// 0x0C TI_DARK_GREEN
        { 0xB0, 0x00, 0xB0, 0x00 },		// 0x0D TI_MAGENTA
        { 0xCC, 0xCC, 0xCC, 0x00 },		// 0x0E TI_GRAY
        { 0xFF, 0xFF, 0xFF, 0x00 },		// 0x0F TI_WHITE
        { 0xFF, 0xFF, 0xFF, 0x00 }		// 0x10 - Text Mode Foreground
    }, {                    
        { 0x00, 0x00, 0x00, 0x00 },		// 0x00 TI_TRANSPARENT - Background
        { 0x00, 0x00, 0x00, 0x00 },		// 0x01 TI_BLACK
        { 0x48, 0x9C, 0x08, 0x00 },		// 0x02 TI_MEDIUM_GREEN *
        { 0x70, 0xBF, 0x88, 0x00 },		// 0x03 TI_LIGHT_GREEN *
        { 0x28, 0x3C, 0x8A, 0x00 },		// 0x04 TI_DARK_BLUE *
        { 0x50, 0x6C, 0xCF, 0x00 },		// 0x05 TI_LIGHT_BLUE *
        { 0xD0, 0x48, 0x00, 0x00 },		// 0x06 TI_DARK_RED *
        { 0x00, 0xCC, 0xFF, 0x00 },		// 0x07 TI_CYAN *
        { 0xD0, 0x58, 0x28, 0x00 },		// 0x08 TI_MEDIUM_RED *
        { 0xFF, 0xA0, 0x40, 0x00 },		// 0x09 TI_LIGHT_RED *
        { 0xFC, 0xF0, 0x50, 0x00 },		// 0x0A TI_DARK_YELLOW *
        { 0xFF, 0xFF, 0x80, 0x00 },		// 0x0B TI_LIGHT_YELLOW *
        { 0x00, 0x80, 0x00, 0x00 },		// 0x0C TI_DARK_GREEN
        { 0xCD, 0x58, 0xCD, 0x00 },		// 0x0D TI_MAGENTA *
        { 0xE0, 0xE0, 0xE0, 0x00 },		// 0x0E TI_GRAY *
        { 0xFF, 0xFF, 0xFF, 0x00 },		// 0x0F TI_WHITE
        { 0xFF, 0xFF, 0xFF, 0x00 }		// 0x10 - Text Mode Foreground
    }
};

static int   joystickIndex [2] = { 0, 1 };
static int   framesOn          = 0;
static int   framesOff         = 0;
static char *diskImage [3];

bool ListJoysticks ( const char *, void * )
{
    FUNCTION_ENTRY ( NULL, "ListJoysticks", true );

    int num = SDL_NumJoysticks ();
    cprintf ( "%i joysticks were found.\n", num );

    if ( num > 0 ) cprintf ( "\nThe names of the joysticks are:\n" );
    for ( int i = 0; i < SDL_NumJoysticks (); i++ ) {
        cprintf ( "  %d) %s\n", i, SDL_JoystickName (i));
    }

    exit ( 0 );

    return false;
}

bool ParseJoystick ( const char *arg, void * )
{
    FUNCTION_ENTRY ( NULL, "ParseJoystick", true );

    arg += strlen ( "joystick" );

    int joy_num = arg [0] - '0';
    if (( joy_num < 1 ) || ( joy_num > 2 ) || ( arg [1] != '=' )) {
        return false;
    }

    int joy_index = arg [3] - '0';
    if (( joy_index < 0 ) || ( joy_index >= SDL_NumJoysticks ())) {
        return false;
    }
    joystickIndex [joy_num] = joy_index;

    return true;
}

bool ParseDisk ( const char *arg, void * )
{
    FUNCTION_ENTRY ( NULL, "ParseDisk", true );

    arg += strlen ( "dsk" );

    int disk = arg [0] - '1';
    if (( disk < 0 ) || ( disk > 2 ) || ( arg [1] != '=' )) {
        return false;
    }

    diskImage [disk] = strdup ( arg + 2 );

    return true;
}

bool ParseFrameRate ( const char *arg, void * )
{
    FUNCTION_ENTRY ( NULL, "ParseFrameRate", true );

    int num = 0, den = 0;

    arg += strlen ( "framerate=" );

    switch ( sscanf ( arg, "%d/%d", &num, &den )) {
        case 1 :
            den = 100;
            break;
        case 2 : 
            break;
        default :
            cprintf ( "Invalid framerate '%s'\n", arg );
            return false;
    }

    if (( num > den ) || ( den == 0 )) {
        cprintf ( "Invalid framerate specified\n" );
        return false;
    }

    return true;
}

bool IsType ( const char *filename, const char *type )
{
    FUNCTION_ENTRY ( NULL, "IsType", true );

    int len = strlen ( filename );
    const char *ptr = filename + len - 4;
    return ( strcmp ( ptr, type ) == 0 ) ? true : false;
}

void PrintUsage ()
{
    FUNCTION_ENTRY ( NULL, "PrintUsage", true );

    cprintf ( "Usage: ti99sim-sdl [options] [cartridge] [image]\n" );
    cprintf ( "\n" );
}

int main ( int argc, char *argv [] )
{
    FUNCTION_ENTRY ( NULL, "main", true );

    int flagSize        = 1;
    bool flagSound      = true;
    bool flagSpeech     = true;
    bool flagFullScreen = false;
    int refreshRate     = 60;

    sOption optList [] = {
        { '4', NULL,                 OPT_VALUE_SET | OPT_SIZE_INT,  2,     &flagSize,       NULL,           "Double width/height window" },
        { 'f', "fullscreen",         OPT_VALUE_SET | OPT_SIZE_BOOL, true,  &flagFullScreen, NULL,           "Fullscreen" },
        { 'q', "nosound",            OPT_VALUE_SET | OPT_SIZE_BOOL, false, &flagSound,      NULL,           "Turn off all sound/speech" },
        {  0,  "nospeech",           OPT_VALUE_SET | OPT_SIZE_BOOL, false, &flagSpeech,     NULL,           "Disable speech synthesis" },
        {  0,  "joystick*n=i",       OPT_NONE,                      0,     NULL,            ParseJoystick,  "Use system joystick i as TI joystick n" },
        {  0,  "dsk*n=filename",     OPT_NONE,                      0,     NULL,            ParseDisk,      "Use 'filename' disk image for DSKn" },
        {  0,  "listjoysticks",      OPT_NONE,                      0,     NULL,            ListJoysticks,  "Print a list of all detected joysticks" },
        {  0,  "NTSC",               OPT_VALUE_SET | OPT_SIZE_INT,  60,    &refreshRate,    NULL,           "Emulate a NTSC display (60Hz)" },
        {  0,  "PAL",                OPT_VALUE_SET | OPT_SIZE_INT,  50,    &refreshRate,    NULL,           "Emulate a PAL display (50Hz)" },
        {  0,  "framerate*={n/d|p}", OPT_NONE,                      0,     NULL,            ParseFrameRate, "Reduce frame rate to fraction n/d or percentage p" }
    };

    // Initialize the SDL library (starts the event loop)
    if ( SDL_Init ( SDL_INIT_NOPARACHUTE | SDL_INIT_VIDEO | SDL_INIT_JOYSTICK ) < 0 ) {
        fprintf ( stderr, "Couldn't initialize SDL: %s\n", SDL_GetError ());
        return -1;
    }

    // Clean up on exit, exit on window close and interrupt
    atexit ( SDL_Quit );

    SDL_WM_SetCaption ( "TI-99/sim", NULL );

    int index = 1;
    index = ParseArgs ( index, argc, argv, SIZE ( optList ), optList );

    SDL_Joystick *joy1 = SDL_JoystickOpen ( joystickIndex [0] );
    SDL_Joystick *joy2 = SDL_JoystickOpen ( joystickIndex [1] );

    VIDEO_SIZE_E size = VIDEO_NORMAL;
    if ( flagSize == 2 ) size = VIDEO_DOUBLE;
    if ( flagFullScreen == ( int ) true ) size = VIDEO_FULLSCREEN;
    if ( flagSound == ( int ) false ) flagSpeech = false;
    cTMS9918A *vdp = new cSdlTMS9918A ( ColorTable [1], size, refreshRate, framesOn, framesOff );
    cTMS9919 *sound = ( flagSound == false ) ? new cTMS9919 () : new cSdlTMS9919 ();
    cTMS5220 *speech = ( flagSpeech == false ) ? NULL : new cTMS5220 ( sound );
    cSdlTI994A computer ( new cCartridge ( LocateFile ( "roms", "TI-994A.ctg" )), vdp, sound, speech );

    cDiskDevice *disk = new cDiskDevice ( LocateFile ( "roms", "ti-disk.ctg" ));
    for ( unsigned i = 0; i < SIZE ( diskImage ); i++ ) {
        char dskName [10];
        sprintf ( dskName, "dsk%d.dsk", i + 1 );
        const char *validName = LocateFile ( "disks", diskImage [i] );
        if ( validName == NULL ) {
            validName = LocateFile ( "disks", dskName );
            if ( validName == NULL ) {
                validName = dskName;
            }
        }
        disk->LoadDisk ( i, validName );
    }
    computer.AddDevice ( disk );

    computer.SetJoystick ( 0, joy1 );
    computer.SetJoystick ( 1, joy2 );

    cCartridge *ctg = NULL;
    while ( index < argc ) {
        if ( IsType ( argv [index], ".ctg" )) {
            if ( ctg != NULL ) {
                computer.RemoveCartridge ( ctg );
                delete ctg;
            }
            ctg = new cCartridge ( LocateFile ( "cartridges", argv [index] ));
            computer.InsertCartridge ( ctg );
        }

        if ( IsType ( argv [index], ".img" )) {
            (( cTI994A & ) computer).LoadImage ( argv [index] );
        }

        index++;
    }

    computer.Run ();

    if ( ctg != NULL ) {
        computer.RemoveCartridge ( ctg );
        delete ctg;
    }

    for ( unsigned i = 0; i < SIZE ( diskImage ); i++ ) {
        if ( diskImage [i] != NULL ) {
            free ( diskImage [i] );
        }
    }

    if ( SDL_JoystickOpened ( joystickIndex [0] )) SDL_JoystickClose ( joy1 );
    if ( SDL_JoystickOpened ( joystickIndex [1] )) SDL_JoystickClose ( joy2 );

    return 0;
}
